// threshold.c
// makes new RGB image by  channel thresholding
// heavily, and in a dirty(?) way, reuses code from resample.c
// sorry,
// Thu Jul 11 06:46:17 EDT 2002  <pac@next.gli.cas.cz>
// Version 1.00


#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>

#define ALPHA 3
#define RED 2
#define GREEN 1
#define BLUE 0


void
usage(void)
{
	fprint(2, "usage: threshold value [imagefile]\n");
	fprint(2, "           jpg -3 testimg.jpg | threshold 128  | page  \n");
	exits("usage");
}

void
thresholdchannel(uchar **in, int off, int iny, uchar **out, int outy, int n, int thres)
{
	int y,  current=255;	
	switch(n){
	case 3:
	case 4: 
		for(y=0; y<outy; y++){
			if (in[y][off]<=thres) current = 0; else current = 255;
			out[y][off] = current;
		}  //for
	break;
	default:
		sysfatal("can't handle  %d-channel images, need 24(32)-bpp depth", n);	
	}  //switch (n)
}


int
max(int a, int b)
{
	if(a > b)
		return a;
	return b;
}

Memimage*
mknewimg(int xsize, int ysize, Memimage *m, int thres)
{
	int i, j, bpl, nchan;
	Memimage *new;
	uchar **oscan, **nscan;

	new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
	if(new == nil)
		sysfatal("can't allocate new image: %r");

	oscan = malloc(Dy(m->r)*sizeof(uchar*));
	nscan = malloc(max(ysize, Dy(m->r))*sizeof(uchar*));
	if(oscan == nil || nscan == nil)
		sysfatal("can't allocate: %r");

	/* unload original image into scan lines */
	bpl = bytesperline(m->r, m->depth);
	for(i=0; i<Dy(m->r); i++){
		oscan[i] = malloc(bpl);
		if(oscan[i] == nil)
			sysfatal("can't allocate: %r");
		j = unloadmemimage(m, Rect(m->r.min.x, m->r.min.y+i, m->r.max.x, m->r.min.y+i+1), oscan[i], bpl);
		if(j != bpl)
			sysfatal("unloadmemimage");
	}

	/* allocate scan lines for destination. we do y first, so need at least Dy(m->r) lines */
	bpl = bytesperline(Rect(0, 0, xsize, Dy(m->r)), m->depth);
	for(i=0; i<max(ysize, Dy(m->r)); i++){
		nscan[i] = malloc(bpl);
		if(nscan[i] == nil)
			sysfatal("can't allocate: %r");
	}

	nchan = m->depth/8;

	/* threshold channel in Y direction*/
	for(i=0; i<xsize; i++)
		for(j=0; j<nchan; j++)
			thresholdchannel(oscan, nchan*i+j, Dy(m->r), nscan, ysize, nchan, thres);

	/* pack data into destination */
	bpl = bytesperline(new->r, m->depth);
	for(i=0; i<ysize; i++){
		j = loadmemimage(new, Rect(0, i, xsize, i+1), nscan[i], bpl);
		if(j != bpl)
			sysfatal("loadmemimage: %r");
	}
	return new;
}

void
main(int argc, char *argv[])
{
	int i, fd, xsize, ysize, thres;
	Rectangle rparam;
	Memimage *m, *new, *t1, *t2;
	char *file;
	ulong tchan;
	char tmp[100];
	double v;

	memimageinit();
	memset(&rparam, 0, sizeof rparam);
	xsize = ysize = 0;

	ARGBEGIN{
	}ARGEND

	file = "<stdin>";
	fd = 0;

	switch (argc) {
	case 2:
		file = argv[1];
		fd = open(file, OREAD);
		if(fd < 0)
			sysfatal("can't open %s: %r", file);
	case 1:
		thres = atoi(argv[0]);
		break;
	default:
		usage();
	}

	m = readmemimage(fd);
	if(m == nil)
		sysfatal("can't read %s: %r", file);

	xsize = Dx(m->r);
	ysize = Dy(m->r);

	new = nil;
	switch(m->chan){

	case RGB24:
	case RGBA32:
	case ARGB32:
	case XRGB32:
		new = mknewimg(xsize, ysize, m, thres);
		break;

	case GREY1:
	case GREY2:
	case GREY4:
	case GREY8:
	case CMAP8:
	case RGB15:
	case RGB16:
		tchan = RGB24;
		goto Convert;

	Convert:
		/* use library to convert to byte-per-chan form, then convert back */
		t1 = allocmemimage(m->r, tchan);
		if(t1 == nil)
			sysfatal("can't allocate temporary image: %r");
		memimagedraw(t1, t1->r, m, m->r.min, nil, ZP);
		t2 = mknewimg(xsize, ysize, t1, thres);
		freememimage(t1);
		new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
		if(new == nil)
			sysfatal("can't allocate new image: %r");
		/* should do error diffusion here */
		memimagedraw(new, new->r, t2, t2->r.min, nil, ZP);
		freememimage(t2);
		break;

	default:
		sysfatal("can't handle channel type %s", chantostr(tmp, m->chan));
	}

	assert(new);
	if(writememimage(1, new) < 0)
		sysfatal("write error on output: %r");

	exits(nil);
}
